@@ -401,6 +401,14 @@
 #define MFD_ALLOW_SEALING 0x0002U
 #endif
 
+#ifndef MFD_NOEXEC_SEAL
+#define MFD_NOEXEC_SEAL 0x0008U
+#endif
+
+#ifndef MFD_EXEC
+#define MFD_EXEC 0x0010U
+#endif
+
 /**
  * BUILD_BUG_ON - break compile if a condition is true.
  * @condition: the condition which the compiler should know is false.
@@ -455,6 +463,14 @@ extern int __build_bug_on_failed;
 #define IFLA_NET_NS_FD 28
 #endif
 
+#ifndef IFLA_PROP_LIST
+#define IFLA_PROP_LIST 52
+#endif
+
+#ifndef IFLA_ALT_IFNAME
+#define IFLA_ALT_IFNAME 53
+#endif
+
 #ifndef IFLA_INFO_KIND
 #define IFLA_INFO_KIND 1
 #endif
@@ -524,6 +540,14 @@ extern int __build_bug_on_failed;
 #define RTM_GETNSID 90
 #endif
 
+#ifndef RTM_NEWLINKPROP
+#define RTM_NEWLINKPROP 108
+#endif
+
+#ifndef RTM_DELLINKPROP
+#define RTM_DELLINKPROP 109
+#endif
+
 #ifndef NLMSG_ERROR
 #define NLMSG_ERROR 0x2
 #endif
